"""Defines lumped elements that should be included in the simulation."""

from __future__ import annotations

from abc import ABC, abstractmethod
from math import isclose
from typing import Annotated, Literal, Optional, Union

import numpy as np
import pydantic.v1 as pd

from tidy3d.components.grid.grid import Grid
from tidy3d.components.medium import PEC2D, Debye, Drude, Lorentz, Medium, Medium2D, PoleResidue
from tidy3d.components.monitor import FieldMonitor
from tidy3d.components.structure import MeshOverrideStructure, Structure
from tidy3d.components.validators import assert_line_or_plane, assert_plane, validate_name_str
from tidy3d.constants import EPSILON_0, FARAD, HENRY, MICROMETER, OHM, fp_eps
from tidy3d.exceptions import ValidationError
from tidy3d.log import log

from .base import Tidy3dBaseModel, cached_property, skip_if_fields_missing
from .geometry.base import Box, ClipOperation, Geometry, GeometryGroup
from .geometry.primitives import Cylinder
from .geometry.utils import (
    SnapBehavior,
    SnapLocation,
    SnappingSpec,
    snap_box_to_grid,
    snap_point_to_grid,
)
from .geometry.utils_2d import increment_float
from .microwave.formulas.circuit_parameters import (
    capacitance_colinear_cylindrical_wire_segments,
    capacitance_rectangular_sheets,
    inductance_straight_rectangular_wire,
    total_inductance_colinear_rectangular_wire_segments,
)
from .types import (
    TYPE_TAG_STR,
    Axis,
    Axis2D,
    Coordinate,
    CoordinateOptional,
    FreqArray,
    LumpDistType,
)
from .viz import PlotParams, plot_params_lumped_element

DEFAULT_LUMPED_ELEMENT_NUM_CELLS = 1
LOSS_FACTOR_INDUCTOR = 1e6


class LumpedElement(Tidy3dBaseModel, ABC):
    """Base class describing the interface all lumped elements obey."""

    name: str = pd.Field(
        ...,
        title="Name",
        description="Unique name for the lumped element.",
        min_length=1,
    )

    num_grid_cells: Optional[pd.PositiveInt] = pd.Field(
        DEFAULT_LUMPED_ELEMENT_NUM_CELLS,
        title="Lumped element grid cells",
        description="Number of mesh grid cells associated with the lumped element along each direction. "
        "Used in generating the suggested list of :class:`.MeshOverrideStructure` objects. "
        "A value of ``None`` will turn off mesh refinement suggestions.",
    )

    enable_snapping_points: bool = pd.Field(
        True,
        title="Snap Grid To Lumped Element",
        description="When enabled, snapping points are automatically generated to snap grids to key "
        "geometric features of the lumped element for more accurate modelling.",
    )

    _name_validator = validate_name_str()

    @cached_property
    def plot_params(self) -> PlotParams:
        """Default parameters for plotting a :class:`.LumpedElement` object."""
        return plot_params_lumped_element

    @abstractmethod
    def to_mesh_overrides(self) -> list[MeshOverrideStructure]:
        """Creates a suggested :class:`.MeshOverrideStructure` list."""

    @abstractmethod
    def to_snapping_points(self) -> list[CoordinateOptional]:
        """Creates a suggested snapping point list."""

    @abstractmethod
    def to_geometry(self) -> Geometry:
        """Converts the :class:`.LumpedElement` object to a :class:`.Geometry`."""

    @abstractmethod
    def to_structure(self, grid: Grid = None) -> Structure:
        """Converts the network portion of the :class:`.LumpedElement` object to a
        :class:`.Structure`."""

    def to_structures(self, grid: Grid = None) -> list[Structure]:
        """Converts the :class:`.LumpedElement` object to a list of :class:`.Structure`
        which are ready to be added to the :class:`.Simulation`"""
        return [self.to_structure(grid)]

    @pd.root_validator(pre=False)
    def _warn_rf_license(cls, values):
        log.warning(
            "ℹ️ ⚠️ RF simulations are subject to new license requirements in the future. You have instantiated at least one RF-specific component.",
            log_once=True,
        )
        return values


class RectangularLumpedElement(LumpedElement, Box):
    """Class representing a rectangular element with zero thickness. A :class:`RectangularLumpedElement`
    is appended to the list of structures in the simulation as a :class:`.Medium2D` with the appropriate
    material properties given their size, voltage axis, and the network they represent."""

    voltage_axis: Axis = pd.Field(
        ...,
        title="Voltage Drop Axis",
        description="Specifies the axis along which the component is oriented and along which the "
        "associated voltage drop will occur. Must be in the plane of the element.",
    )

    snap_perimeter_to_grid: bool = pd.Field(
        True,
        title="Snap Perimeter to Grid",
        description="When enabled, the perimeter of the lumped element is snapped to the simulation grid, "
        "which improves accuracy when the number of grid cells is low within the element. Sides of the element "
        "perpendicular to the ``voltage_axis`` are snapped to grid boundaries, while the sides parallel to the "
        "``voltage_axis`` are snapped to grid centers. Lumped elements are always snapped to the nearest grid "
        "boundary along their ``normal_axis``, regardless of this option.",
    )

    _line_plane_validator = assert_line_or_plane()

    @cached_property
    def normal_axis(self):
        """Normal axis of the lumped element, which is the axis where the element has zero size."""
        return self.size.index(0.0)

    @cached_property
    def lateral_axis(self):
        """Lateral axis of the lumped element."""
        return 3 - self.voltage_axis - self.normal_axis

    @cached_property
    def _voltage_axis_2d(self) -> Axis2D:
        """Returns the voltage axis using the in-plane dimensions used by :class:`.Medium2D`."""
        if self.normal_axis > self.voltage_axis:
            return self.voltage_axis
        return self.voltage_axis - 1

    @cached_property
    def _snapping_spec(self) -> SnappingSpec:
        """Returns the snapping behavior for each dimension of the lumped element.

        Note
        ----

        Snapping the lumped element is needed for accuracy, since in many cases staircasing
        will be used instead of subpixel averaging, e.g., when there are many different media
        in close proximity to the lumped element. The equivalent media produced by the
        lumped element is usually quite large in magnitude and depends directly on the physical
        dimensions of the lumped element. As a result, we need to ensure that staircasing will
        lead to an accurate representation of the lumped element. We make sure to create a
        :class:`.Box` aligned with the Yee grid that fully encloses the
        electric field component which is parallel to the ``voltage_axis``.
        """

        snap_location = [SnapLocation.Boundary] * 3
        snap_behavior = [SnapBehavior.Closest] * 3
        snap_location[self.lateral_axis] = SnapLocation.Center
        snap_behavior[self.lateral_axis] = SnapBehavior.Expand
        return SnappingSpec(location=snap_location, behavior=snap_behavior)

    def to_mesh_overrides(self) -> list[MeshOverrideStructure]:
        """Creates a suggested :class:`.MeshOverrideStructure` list for mesh refinement both on the
        plane of lumped element, and along normal axis. In the normal direction, we'll make sure there
        are at least 2 cell layers above and below whose size is the same as the in-plane cell
        size in the override region.
        """

        if self.num_grid_cells is None:
            return []
        dl = self.size[self.voltage_axis] / self.num_grid_cells
        override_size = list(self.size)
        override_size[self.normal_axis] = 4 * dl
        return [
            MeshOverrideStructure(
                geometry=Box(center=self.center, size=override_size),
                dl=(dl, dl, dl),
                shadow=False,
                priority=-1,
            )
        ]

    def to_snapping_points(self) -> list[CoordinateOptional]:
        """Creates a suggested snapping point list to ensure that the element is aligned with a grid
        boundary in the normal direction, and the endpoints aligned with grids in the voltage axis.
        """

        if not self.enable_snapping_points:
            return []
        # normal axis
        snapping_points = [
            Geometry.unpop_axis(self.center[self.normal_axis], (None, None), axis=self.normal_axis)
        ]
        # also snap along voltage axis
        for bound_coord in self.bounds:
            snapping_points.append(
                Geometry.unpop_axis(
                    bound_coord[self.voltage_axis], (None, None), axis=self.voltage_axis
                )
            )
        return snapping_points

    def to_geometry(self, grid: Grid = None) -> Box:
        """Converts the :class:`RectangularLumpedElement` object to a :class:`.Box`."""
        box = Box(size=self.size, center=self.center)
        if grid and self.snap_perimeter_to_grid:
            return snap_box_to_grid(grid, box, self._snapping_spec)
        return box

    def _admittance_transfer_function_scaling(self, box: Box = None) -> float:
        """The admittance transfer function of the network needs to be scaled depending on the dimensions
        of the lumped element. The scaling emulates adding networks with equal admittances in series and
        parallel, and is needed when distributing the network over a finite volume.

        Note
        ----

        The lumped element models the relationship I = Y*V, where I is the current, Y is the admittance,
        and V is the voltage. Assume the ``voltage_axis`` is aligned with the z axis, and dx, dy, and dz
        represent the size of the lumped element. The voltage can be related to electric field by V = dz*Ez.
        Likewise, the current can be related to the current density by I = dx*dy*Jz. Then, the current
        density and electric field within the lumped element can be related to each other by
        Jz = dz/(dx*dy)*Y*Ez. As a result, an equivalent medium needs to be created with a complex conductivity
        that is equal to dz/(dx*dy)*Y. Note that the thickness along the ``normal_axis`` is taken into account
        once the 2D medium is converted into a volumetric object.
        """
        size = self.size
        if box:
            size = box.size
        size_voltage = size[self.voltage_axis]
        size_lateral = size[self.lateral_axis]
        # The final scaling along the normal axis is applied when the resulting 2D medium is averaged with the background media.
        return size_voltage / size_lateral

    def to_monitor(self, freqs: FreqArray) -> FieldMonitor:
        """Creates a field monitor that can be added to the simulation, which records field data
        that can be used to later compute voltage and current flowing through the element.
        """

        center = list(self.center)
        # Size of monitor needs to be nonzero along the normal axis so that the magnetic field on
        # both sides of the sheet will be available
        mon_size = list(self.size)
        mon_size[self.normal_axis] = 2 * (
            increment_float(center[self.normal_axis], 1.0) - center[self.normal_axis]
        )

        e_component = "xyz"[self.voltage_axis]
        h1_component = "xyz"[self.lateral_axis]
        h2_component = "xyz"[self.normal_axis]
        # Create a voltage monitor
        return FieldMonitor(
            center=center,
            size=mon_size,
            freqs=freqs,
            fields=[f"E{e_component}", f"H{h1_component}", f"H{h2_component}"],
            name=self.monitor_name,
            colocate=False,
        )

    @cached_property
    def monitor_name(self):
        return f"{self.name}_monitor"

    @pd.validator("voltage_axis", always=True)
    @skip_if_fields_missing(["name", "size"])
    def _voltage_axis_in_plane(cls, val, values):
        """Ensure voltage drop axis is in the plane of the lumped element."""
        name = values.get("name")
        size = values.get("size")
        if size.count(0.0) == 1 and size.index(0.0) == val:
            # if not planar, then a separate validator should be triggered, not this one
            raise ValidationError(
                f"'voltage_axis' must be in the plane of lumped element '{name}'."
            )
        return val


class LumpedResistor(RectangularLumpedElement):
    """Class representing a rectangular lumped resistor. Lumped resistors are appended to the list
    of structures in the simulation as :class:`Medium2D` with the appropriate conductivity given
    their size and voltage axis."""

    resistance: pd.PositiveFloat = pd.Field(
        ...,
        title="Resistance",
        description="Resistance value in ohms.",
        unit=OHM,
    )

    def _sheet_conductance(self, box: Box = None):
        """Effective sheet conductance."""
        return self._admittance_transfer_function_scaling(box) / self.resistance

    def to_structure(self, grid: Grid = None) -> Structure:
        """Converts the :class:`LumpedResistor` object to a :class:`.Structure`
        ready to be added to the :class:`.Simulation`"""
        box = self.to_geometry(grid=grid)
        conductivity = self._sheet_conductance(box)
        components_2d = ["ss", "tt"]
        voltage_component = components_2d.pop(self._voltage_axis_2d)
        other_component = components_2d[0]
        medium_dict = {
            voltage_component: Medium(conductivity=conductivity),
            other_component: Medium(permittivity=1),
        }
        return Structure(
            geometry=box,
            medium=Medium2D(**medium_dict),
        )

    _plane_validator = assert_plane()


class CoaxialLumpedResistor(LumpedElement):
    """Class representing a coaxial lumped resistor. Lumped resistors are appended to the list of
    structures in the simulation as :class:`Medium2D` with the appropriate conductivity given their
    size and geometry."""

    resistance: pd.PositiveFloat = pd.Field(
        ...,
        title="Resistance",
        description="Resistance value in ohms.",
        unit=OHM,
    )

    center: Coordinate = pd.Field(
        (0.0, 0.0, 0.0),
        title="Center",
        description="Center of object in x, y, and z.",
        units=MICROMETER,
    )

    outer_diameter: pd.PositiveFloat = pd.Field(
        ...,
        title="Outer Diameter",
        description="Diameter of the outer concentric circle.",
        units=MICROMETER,
    )

    inner_diameter: pd.PositiveFloat = pd.Field(
        ...,
        title="Inner Diameter",
        description="Diameter of the inner concentric circle.",
        units=MICROMETER,
    )

    normal_axis: Axis = pd.Field(
        ...,
        title="Normal Axis",
        description="Specifies the normal axis, which defines "
        "the orientation of the circles making up the coaxial lumped element.",
    )

    def to_snapping_points(self) -> list[CoordinateOptional]:
        """Creates a suggested snapping point list to ensure that the element is aligned with a grid
        boundary in the normal direction."""
        if not self.enable_snapping_points:
            return []

        return [
            Geometry.unpop_axis(self.center[self.normal_axis], (None, None), axis=self.normal_axis)
        ]

    def to_mesh_overrides(self) -> list[MeshOverrideStructure]:
        """Creates a suggested :class:`.MeshOverrideStructure` list for mesh refinement both on the
        plane of lumped element, and along normal axis. In the normal direction, we'll make sure there
        are at least 2 cell layers above and below whose size is half of the in-plane cell
        size in the override region.
        """

        if self.num_grid_cells is None:
            return []
        # Make sure the number of grid cells between inner and outer radius is `self.num_grid_cells`
        dl = (self.outer_diameter - self.inner_diameter) / self.num_grid_cells / 2
        override_dl = Geometry.unpop_axis(dl / 2, (dl, dl), axis=self.normal_axis)
        override_size = Geometry.unpop_axis(
            dl * 2, (self.outer_diameter, self.outer_diameter), axis=self.normal_axis
        )
        return [
            MeshOverrideStructure(
                geometry=Box(center=self.center, size=override_size),
                dl=override_dl,
                shadow=False,
                priority=-1,
            )
        ]

    @pd.validator("center", always=True)
    def _center_not_inf(cls, val):
        """Make sure center is not infinitiy."""
        if any(np.isinf(v) for v in val):
            raise ValidationError("'center' can not contain 'td.inf' terms.")
        return val

    @pd.validator("inner_diameter", always=True)
    @skip_if_fields_missing(["outer_diameter"])
    def _ensure_inner_diameter_is_smaller(cls, val, values):
        """Ensures that the inner diameter is smaller than the outer diameter, so that the final shape is an annulus."""
        outer_diameter = values.get("outer_diameter")
        if val >= outer_diameter:
            raise ValidationError(
                f"The 'inner_diameter' {val} of a coaxial lumped element must be less than its 'outer_diameter' {outer_diameter}."
            )
        return val

    @cached_property
    def _sheet_conductance(self):
        """Effective sheet conductance for a coaxial resistor."""
        rout = self.outer_diameter / 2
        rin = self.inner_diameter / 2
        return 1 / (2 * np.pi * self.resistance) * (np.log(rout / rin))

    def to_structure(self, grid: Grid = None) -> Structure:
        """Converts the :class:`CoaxialLumpedResistor` object to a :class:`.Structure`
        ready to be added to the :class:`.Simulation`"""
        conductivity = self._sheet_conductance
        medium_dict = {
            "tt": Medium(conductivity=conductivity),
            "ss": Medium(conductivity=conductivity),
        }
        return Structure(
            geometry=self.to_geometry(grid),
            medium=Medium2D(**medium_dict),
        )

    def to_geometry(self, grid: Grid = None) -> ClipOperation:
        """Converts the :class:`CoaxialLumpedResistor` object to a :class:`Geometry`."""
        rout = self.outer_diameter / 2
        rin = self.inner_diameter / 2
        disk_out = Cylinder(axis=self.normal_axis, radius=rout, length=0, center=self.center)
        disk_in = Cylinder(axis=self.normal_axis, radius=rin, length=0, center=self.center)
        annulus = ClipOperation(operation="difference", geometry_a=disk_out, geometry_b=disk_in)
        return annulus

    @cached_property
    def geometry(self) -> ClipOperation:
        """Alias for ``to_geometry`` that ignores the grid and allows :class:`CoaxialLumpedResistor`
        to behave like a :class:`.Structure`.

        Returns
        -------
        ClipOperation
            The annulus describing the coaxial lumped resistor.
        """
        return self.to_geometry()


class NetworkConversions(Tidy3dBaseModel):
    """Helper functionality for directly computing complex conductivity and permittivities using
    equations in _`[1]`. Useful for testing the direct translations of lumped network parameters into
    an equivalent PoleResidue medium.

    Notes
    -----

        This implementation follows a similar approach as _`[1]` with a couple small differences. Instead of
        scaling the complex conductivity by the size of a single grid cell, we later scale the quantities by the
        size of the lumped element in the FDTD simulation. In many cases, we will assume the time step is small,
        so that the complex conductivity can be expressed more simply as a rational expression.

        **References**

        .. [1]  J. A. Pereda, F. Alimenti, P. Mezzanotte, L. Roselli and R. Sorrentino, "A new algorithm
                for the incorporation of arbitrary linear lumped networks into FDTD simulators," IEEE
                Trans. Microw. Theory Tech., vol. 47, no. 6, pp. 943-949, Jun. 1999.
    """

    @staticmethod
    def complex_conductivity(a: tuple[float, ...], b: tuple[float, ...], freqs: np.ndarray):
        """Returns the equivalent conductivity of the lumped network over the range of frequencies
        provided in ``freqs`` using the expression in _`[1]`.

        Parameters
        ----------
        a : tuple[float, ...]
            Coefficients of the numerator polynomial
        b : tuple[float, ...]
            Coefficients of the denominator polynomial.
        freqs: np.ndarray
            Frequencies at which to evaluate model.

        Returns
        -------
        np.ndarray
            The resulting complex conductivity.
        """

        # This is the original term from [1], instead we use the limiting case of dt -> 0.
        # After time-discretization, the PoleResidue medium should model the original term.
        # K_tan = -1j * (2 / dt) * np.tan(2 * np.pi * freqs * dt / 2)
        K_tan = -1j * 2 * np.pi * freqs
        numer = 0
        denom = 0
        for a_m, m in zip(a, range(len(a))):
            numer += a_m * K_tan ** (m)
        for b_m, m in zip(b, range(len(b))):
            denom += b_m * K_tan ** (m)
        # We do not include the scaling factor associated with the cell size, since we will
        # distribute the network over more than one cell.
        return numer / denom

    @staticmethod
    def complex_permittivity(a: tuple[float, ...], b: tuple[float, ...], freqs: np.ndarray):
        """
        Returns an equivalent complex permittivity of the lumped network over the range of frequencies
        provided in ``freqs`` using the expression in _`[1]`. The result needs to be combined with a
        :math:`\\epsilon_\\infty`, e.g., 1 or the existing background medium, before being added to an
        FDTD simulation.

        Parameters
        ----------
        a : tuple[float, ...]
            Coefficients of the numerator polynomial
        b : tuple[float, ...]
            Coefficients of the denominator polynomial.
        freqs: np.ndarray
            Frequencies at which to evaluate model.

        Returns
        -------
        np.ndarray
            The equivalent frequency-dependent portion of the electric permittivity.
        """

        # For fitting with a pole-residue model, we provide a convenience function for
        # converting the complex conductivity to a complex permittivity.
        sigma = NetworkConversions.complex_conductivity(a, b, freqs)
        return 1j * sigma / (2 * np.pi * freqs * EPSILON_0)

    @pd.root_validator(pre=False)
    def _warn_rf_license(cls, values):
        log.warning(
            "ℹ️ ⚠️ RF simulations are subject to new license requirements in the future. You have instantiated at least one RF-specific component.",
            log_once=True,
        )
        return values


class RLCNetwork(Tidy3dBaseModel):
    """Class for representing a simple network consisting of a resistor, capacitor, and inductor.
    Provides additional functionality for representing the network as an equivalent medium.

    Notes
    -----

        Implementation is based on the equivalent medium introduced by _`[1]`.

        **References**

        .. [1]  J. A. Pereda, F. Alimenti, P. Mezzanotte, L. Roselli and R. Sorrentino, "A new algorithm
                for the incorporation of arbitrary linear lumped networks into FDTD simulators," IEEE
                Trans. Microw. Theory Tech., vol. 47, no. 6, pp. 943-949, Jun. 1999.

    Example
    -------
    >>> RL_series = RLCNetwork(resistance=75,
    ...                        inductance=1e-9,
    ...                        network_topology="series"
    ...                       ) # doctest: +SKIP

    """

    resistance: Optional[pd.PositiveFloat] = pd.Field(
        None,
        title="Resistance",
        description="Resistance value in ohms.",
        unit=OHM,
    )

    capacitance: Optional[pd.PositiveFloat] = pd.Field(
        None,
        title="Capacitance",
        description="Capacitance value in farads.",
        unit=FARAD,
    )

    inductance: Optional[pd.PositiveFloat] = pd.Field(
        None,
        title="Inductance",
        description="Inductance value in henrys.",
        unit=HENRY,
    )

    network_topology: Literal["series", "parallel"] = pd.Field(
        "series",
        title="Network Topology",
        description="Describes whether network elements are connected in ``series`` or ``parallel``.",
    )

    @cached_property
    def _number_network_elements(self) -> pd.PositiveInt:
        num_elements = 0
        if self.resistance:
            num_elements += 1
        if self.capacitance:
            num_elements += 1
        if self.inductance:
            num_elements += 1
        return num_elements

    @cached_property
    def _as_admittance_function(self) -> tuple[tuple[float, ...], tuple[float, ...]]:
        """Converts the :class:`RLCNetwork` instance into a rational expression representing the
        admittance of the network in the Laplace domain.
        """
        R = self.resistance
        C = self.capacitance
        L = self.inductance
        if self._number_network_elements == 1:
            # Lumped element is simply a resistor, inductor, or capacitor.
            if R:
                return ((1, 0), (R, 0))
            if C:
                return ((0, C), (1, 0))
            if L:
                return ((1, 0), (0, L))
        if self.network_topology == "series":
            return RLCNetwork._to_series_network_transfer_function(R, L, C)
        return RLCNetwork._to_parallel_network_transfer_function(R, L, C)

    @staticmethod
    def _to_series_network_transfer_function(
        R: float, L: float, C: float
    ) -> tuple[tuple[float, ...], tuple[float, ...]]:
        """Converts the RLC series network to a rational expression representing the
        admittance of the network in the Laplace domain.
        """
        if R and C and L:
            # RLC series network
            a = (0, C, 0)
            b = (1, R * C, L * C)
        elif R and C:
            # RC series network
            a = (0, C)
            b = (1, R * C)
        elif R and L:
            # RL series network
            a = (1, 0)
            b = (R, L)
        else:
            # LC series network
            a = (0, C, 0)
            b = (1, 0, L * C)
        return (a, b)

    @staticmethod
    def _to_parallel_network_transfer_function(
        R: float, L: float, C: float
    ) -> tuple[tuple[float, ...], tuple[float, ...]]:
        """Converts the RLC parallel network to a rational expression representing the
        admittance of the network in the Laplace domain.
        """
        if R and C and L:
            # RLC parallel network
            a = (R, L, R * L * C)
            b = (0, R * L, 0)
        elif R and C:
            # RC parallel network
            a = (1, R * C)
            b = (R, 0)
        elif R and L:
            # RL parallel network
            a = (R, L)
            b = (0, R * L)
        else:
            # LC parallel network
            a = (1, 0, L * C)
            b = (0, L, 0)
        return (a, b)

    def _to_medium(self, scaling_factor: float) -> PoleResidue:
        """Converts the :class:`RLCNetwork` model directly into a :class:`PoleResidue` model
        with proper scaling depending on the lumped element's dimensions."""
        R = self.resistance
        C = self.capacitance
        L = self.inductance

        # eps_infinity is set to 1, simply to avoid validation errors.
        # The final equivalent medium to be added to the simulation needs to be combined with the
        # background medium, where this value of 1 will be ignored and the value of the background
        # medium will be taken.
        if self._number_network_elements == 1:
            # Lumped element is simply a resistor, inductor, or capacitor.
            if R:
                # Technically zeroth order network
                med = Medium(permittivity=1, conductivity=(scaling_factor / R))
                return PoleResidue.from_medium(med)
            if C:
                med = Medium(permittivity=1 + scaling_factor * C / EPSILON_0, conductivity=0)
                return PoleResidue.from_medium(med)
            if L:
                # TODO would be nice to be able to set the damping term exactly to 0
                fi = np.sqrt(scaling_factor / (EPSILON_0 * (2 * np.pi) ** 2 * L))
                # Choose a relatively small value for damping term
                di = fi / LOSS_FACTOR_INDUCTOR
                med = Drude(eps_inf=1.0, coeffs=[(fi, di)])
                return med.pole_residue
        elif self.network_topology == "series":
            result_medium = RLCNetwork._series_network_to_equivalent_medium(scaling_factor, R, L, C)
            return result_medium
        else:
            result_medium = RLCNetwork._parallel_network_to_equivalent_medium(
                scaling_factor, R, L, C
            )
            return result_medium

    @staticmethod
    def _series_network_to_equivalent_medium(
        admittance_scaling_factor: float, R: float, L: float, C: float
    ) -> PoleResidue:
        """Converts the RLC series network directly to an equivalent medium."""
        if R and L and C:
            # RLC series
            delta_eps = admittance_scaling_factor * C / EPSILON_0
            di = R / (4 * np.pi * L)
            fi = np.sqrt(1 / ((2 * np.pi) ** 2 * L * C))
            med = Lorentz(eps_inf=1.0, coeffs=[(delta_eps, fi, di)])
            return med
        if R and C:
            # RC series
            delta_eps = admittance_scaling_factor * C / EPSILON_0
            tau = 2 * np.pi * R * C
            med = Debye(eps_inf=1.0, coeffs=[(delta_eps, tau)])
            return med.pole_residue
        if R and L:
            # RL series
            fi = np.sqrt(admittance_scaling_factor / (EPSILON_0 * (2 * np.pi) ** 2 * L))
            di = R / (2 * np.pi * L)
            med = Drude(eps_inf=1.0, coeffs=[(fi, di)])
            return med.pole_residue
        # LC series
        delta_eps = admittance_scaling_factor * C / EPSILON_0
        di = 0
        fi = np.sqrt(1 / ((2 * np.pi) ** 2 * L * C))
        med = Lorentz(eps_inf=1.0, coeffs=[(delta_eps, fi, di)])
        return med

    @staticmethod
    def _parallel_network_to_equivalent_medium(
        admittance_scaling_factor: float, R: float, L: float, C: float
    ) -> PoleResidue:
        """Converts the RLC parallel network directly to an equivalent medium."""

        def combine_equivalent_medium_in_parallel(first: PoleResidue, second: PoleResidue):
            """Helper for combining equivalent media when the network elements are in the 'parallel'
            configuration. A similar operation cannot be done for the 'series' topology."""
            eps_inf = 1.0 + (first.eps_inf - 1) + (second.eps_inf - 1)
            poles = first.poles + second.poles
            return PoleResidue(eps_inf=eps_inf, poles=poles)

        result_medium = PoleResidue(eps_inf=1, poles=[])
        if R:
            # Add R in parallel
            med = PoleResidue.from_medium(
                Medium(permittivity=1, conductivity=(admittance_scaling_factor / R))
            )
            result_medium = combine_equivalent_medium_in_parallel(result_medium, med)
        if C:
            # C in parallel
            med = PoleResidue.from_medium(
                Medium(
                    permittivity=1 + admittance_scaling_factor * C / EPSILON_0,
                    conductivity=0,
                )
            )
            result_medium = combine_equivalent_medium_in_parallel(result_medium, med)
        if L:
            # L in parallel
            # TODO would be nice to be able to set the damping term exactly to 0
            fi = np.sqrt(admittance_scaling_factor / (EPSILON_0 * (2 * np.pi) ** 2 * L))
            # Choose a relatively small value for damping term
            di = fi / LOSS_FACTOR_INDUCTOR
            med = Drude(eps_inf=1.0, coeffs=[(fi, di)]).pole_residue
            result_medium = combine_equivalent_medium_in_parallel(med, result_medium)
        return result_medium

    @pd.validator("inductance", always=True)
    @skip_if_fields_missing(["resistance", "capacitance"])
    def _validate_single_element(cls, val, values):
        """At least one element should be defined."""
        resistance = values.get("resistance")
        capacitance = values.get("capacitance")
        all_items_are_none = all(item is None for item in [resistance, capacitance, val])
        if all_items_are_none:
            raise ValueError("At least one element must be defined in the 'RLCNetwork'.")
        return val

    @pd.root_validator(pre=False)
    def _warn_rf_license(cls, values):
        log.warning(
            "ℹ️ ⚠️ RF simulations are subject to new license requirements in the future. You have instantiated at least one RF-specific component.",
            log_once=True,
        )
        return values


class AdmittanceNetwork(Tidy3dBaseModel):
    """Class for representing a network consisting of an arbitrary number of resistors,
    capacitors, and inductors. The network is represented in the Laplace domain
    as an admittance function. Provides additional functionality for representing the network
    as an equivalent medium.

    Notes
    -----

        The network is described by the supplied coefficients as an admittance function that relates
        voltage to the current in the Laplace domain and is equivalent to a frequency-dependent
        complex conductivity :math:`\\sigma(\\omega)`.

        .. math::
            I(s) = Y(s)V(s)

        .. math::
            Y(s) = \\frac{a_0 + a_1 s + \\dots + a_M s^M}{b_0 + b_1 s + \\dots + b_N s^N}

        An equivalent :class:`.PoleResidue` medium is constructed using an equivalent frequency-dependent
        complex permittivity defined as

        .. math::
            \\epsilon(s) = \\epsilon_\\infty - \\frac{\\Delta}{\\epsilon_0 s}
            \\frac{a_0 + a_1 s + \\dots + a_M s^M}{b_0 + b_1 s + \\dots + b_N s^N}.

        The admittance is scaled depending on the geometric properties of the lumped element by
        the scaling factor :math:`\\Delta`. Implementation is based on the equivalent medium introduced
        by _`[1]`.

        **References**

        .. [1]  J. A. Pereda, F. Alimenti, P. Mezzanotte, L. Roselli and R. Sorrentino, "A new algorithm
                for the incorporation of arbitrary linear lumped networks into FDTD simulators," IEEE
                Trans. Microw. Theory Tech., vol. 47, no. 6, pp. 943-949, Jun. 1999.

    Example
    -------
    >>> R = 50
    >>> C = 1e-12
    >>> a = (1, R * C) # Coefficients for an RC parallel network
    >>> b = (R, 0)
    >>> RC_parallel = AdmittanceNetwork(a=a,
    ...                                 b=b
    ...               ) # doctest: +SKIP

    """

    a: tuple[pd.NonNegativeFloat, ...] = pd.Field(
        ...,
        title="Numerator Coefficients",
        description="A ``tuple`` of floats describing the coefficients of the numerator polynomial. "
        "The length of the ``tuple`` is equal to the order of the network.",
    )

    b: tuple[pd.NonNegativeFloat, ...] = pd.Field(
        ...,
        title="Denominator Coefficients",
        description="A ``tuple`` of floats describing the coefficients of the denomiator polynomial. "
        "The length of the ``tuple`` is equal to the order of the network.",
    )

    def _to_medium(self, scaling_factor: float) -> PoleResidue:
        """Converts the :class:`AdmittanceNetwork` model directly into a :class:`PoleResidue` model
        with proper scaling depending on the lumped element's dimensions."""
        a = np.array(self.a) * scaling_factor
        b = np.array(self.b)
        return PoleResidue.from_admittance_coeffs(a, b)

    @cached_property
    def _as_admittance_function(self) -> tuple[tuple[float, ...], tuple[float, ...]]:
        """Converts the :class:`AdmittanceNetwork` instance into a rational expression representing the
        admittance of the network in the Laplace domain.
        """
        return (self.a, self.b)

    @pd.root_validator(pre=False)
    def _warn_rf_license(cls, values):
        log.warning(
            "ℹ️ ⚠️ RF simulations are subject to new license requirements in the future. You have instantiated at least one RF-specific component.",
            log_once=True,
        )
        return values


class LinearLumpedElement(RectangularLumpedElement):
    """Lumped element representing a network consisting of resistors, capacitors, and inductors.



    Notes
    -----

        Implementation is based on the equivalent medium introduced by _`[1]`.

        **References**

        .. [1]  J. A. Pereda, F. Alimenti, P. Mezzanotte, L. Roselli and R. Sorrentino, "A new algorithm
                for the incorporation of arbitrary linear lumped networks into FDTD simulators," IEEE
                Trans. Microw. Theory Tech., vol. 47, no. 6, pp. 943-949, Jun. 1999.

    Example
    -------
    >>> RL_series = RLCNetwork(resistance=75,
    ...                        inductance=1e-9,
    ...                        network_topology="series"
    ...             ) # doctest: +SKIP
    >>> linear_element = LinearLumpedElement(
    ...                         center=[0, 0, 0],
    ...                         size=[2, 0, 3],
    ...                         voltage_axis=0,
    ...                         network=RL_series,
    ...                         name="LumpedRL"
    ...                   ) # doctest: +SKIP


    See Also
    --------

    **Notebooks:**
        * `Using lumped elements in Tidy3D simulations <../../notebooks/LinearLumpedElements.html>`_
    """

    network: Union[RLCNetwork, AdmittanceNetwork] = pd.Field(
        ...,
        title="Network",
        description="The linear element produces an equivalent medium that emulates the "
        "voltage-current relationship described by the ``network`` field.",
        discriminator=TYPE_TAG_STR,
    )

    dist_type: LumpDistType = pd.Field(
        "on",
        title="Distribute Type",
        description="Switches between the different methods for distributing the lumped element over "
        "the grid.",
    )
    """
    An advanced feature for :class:`LinearLumpedElement` is the ability to choose different methods
    for distributing the network portion over the the Yee grid. When set to ``on``, the network
    portion of the lumped element is distributed across the entirety of the lumped element's bounding
    box. When set to ``off``, the network portion of the lumped element is restricted to one cell and
    PEC connections are used to connect the network cell to the edges of the lumped element. A third
    option exists ``laterally_only``, where the network portion is only distributed along the lateral
    axis of the lumped element.

    When using a :attr:`dist_type` other than ``on`` additional parasitic network elements are
    introduced, see below. Thin connections lead to a higher inductance, while wide connections
    lead to a higher parasitic capacitance. Follow the link to the associated notebook for an example
    of using this field.

    .. image:: ../../_static/img/lumped_dist_type.png
        :width: 50%

    See Also
    --------
    **Notebooks:**
        * `Using lumped elements in Tidy3D simulations <../../notebooks/LinearLumpedElements.html>`_
    """

    def _create_box_for_network(self, grid: Grid) -> Box:
        """Creates a box for the network portion of the lumped element, where the equivalent
        pole residue medium will be added.
        """
        # Snap center to closest electric field position
        snap_location = 3 * [SnapLocation.Boundary]
        snap_location[self.voltage_axis] = SnapLocation.Center
        cell_center = list(snap_point_to_grid(grid, self.center, snap_location))
        size = [0, 0, 0]

        if self.dist_type != "off" and self.size[self.lateral_axis] != 0:
            cell_center[self.lateral_axis] = self.center[self.lateral_axis]
            size[self.lateral_axis] = self.size[self.lateral_axis]
        if self.dist_type == "on":
            cell_center[self.voltage_axis] = self.center[self.voltage_axis]
            size[self.voltage_axis] = self.size[self.voltage_axis]

        cell_box = Box(center=cell_center, size=size)

        snap_spec = self._snapping_spec
        # Expand from zero size along the voltage and lateral axes
        if size[self.voltage_axis] == 0:
            behavior = list(snap_spec.behavior)
            behavior[self.voltage_axis] = SnapBehavior.Expand
            snap_spec = snap_spec.updated_copy(behavior=behavior)

        return snap_box_to_grid(grid, cell_box, snap_spec=snap_spec)

    def _create_connection_boxes(
        self, cell_box: Box, grid: Grid
    ) -> tuple[Optional[Box], Optional[Box]]:
        """Creates PEC structures that connect the network portion of the lumped element to the
        boundaries of the lumped element.
        """
        element_box = self.to_geometry(grid)
        element_min, element_max = map(list, element_box.bounds)
        cell_min, cell_max = cell_box.bounds

        top_min = list(element_min)
        top_min[self.voltage_axis] = cell_max[self.voltage_axis]
        bottom_max = list(element_max)
        bottom_max[self.voltage_axis] = cell_min[self.voltage_axis]

        # Create "wires" if the size is 0 along the lateral axis
        if isclose(self.size[self.lateral_axis], 0, rel_tol=fp_eps, abs_tol=fp_eps):
            lateral_center = cell_box.center[self.lateral_axis]
            width = max(fp_eps, fp_eps * abs(lateral_center))
            top_min[self.lateral_axis] = lateral_center - width
            element_max[self.lateral_axis] = lateral_center + width
            element_min[self.lateral_axis] = lateral_center - width
            bottom_max[self.lateral_axis] = lateral_center + width

        top_box = Box.from_bounds(top_min, element_max)
        bottom_box = Box.from_bounds(element_min, bottom_max)

        if top_box.size[self.voltage_axis] == 0:
            top_box = None
        if bottom_box.size[self.voltage_axis] == 0:
            bottom_box = None
        return (bottom_box, top_box)

    def to_structure(self, grid) -> Structure:
        """Converts the :class:`LinearLumpedElement` object to a :class:`.Structure`,
        which enforces the desired voltage-current relationship across one or more grid cells."""

        cell_box = self._create_box_for_network(grid)
        medium_scaling_factor = self._admittance_transfer_function_scaling(cell_box)
        medium = self.network._to_medium(medium_scaling_factor)
        components_2d = ["ss", "tt"]
        voltage_component = components_2d.pop(self._voltage_axis_2d)
        other_component = components_2d[0]
        medium_dict = {
            voltage_component: medium,
            other_component: Medium(permittivity=1),
        }
        return Structure(
            geometry=cell_box,
            medium=Medium2D(**medium_dict),
        )

    def to_PEC_connection(self, grid) -> Optional[Structure]:
        """Converts the :class:`LinearLumpedElement` object to a :class:`.Structure`,
        representing any PEC connections.
        """

        if self.dist_type != "on":
            cell_box = self._create_box_for_network(grid)
            connections = self._create_connection_boxes(cell_box, grid)
            connections_filtered = [
                connection for connection in connections if connection is not None
            ]
            if connections_filtered:
                connection_group = GeometryGroup(geometries=connections_filtered)
                structures = Structure(
                    geometry=connection_group,
                    medium=PEC2D,
                )

                return structures

        return None

    def to_structures(self, grid: Grid) -> list[Structure]:
        """Converts the :class:`.LinearLumpedElement` object to a list of :class:`.Structure`
        which are ready to be added to the :class:`.Simulation`"""
        PEC_connection = self.to_PEC_connection(grid)
        structures = []
        if PEC_connection is not None:
            structures.append(PEC_connection)
        structures.append(self.to_structure(grid))
        return structures

    def estimate_parasitic_elements(self, grid: Grid) -> tuple[float, float]:
        """Provides an estimate for the parasitic inductance and capacitance associated with the
        connections. These wire or sheet connections are used when the lumped element is not
        distributed over the voltage axis.

        Notes
        -----
        These estimates for parasitic inductance and capacitance are approximate and may be inaccurate
        in some cases. However, the formulas used should be accurate in the important regime where
        the true values for inductance and capacitance are large. For example, the estimate for capacitance
        will be more accurate for wide elements discretized with a high resolution grid.

        Returns
        -------
        tuple[float, float]
            A tuple containing the parasitic series inductance and parasitic shunt capacitance, respectively.
        """

        if self.dist_type == "on":
            # When connections are not used there is no associated parasitic inductance or capacitance.
            # Note that there is still a small parasitic inductance due to the finite length of the
            # lumped element itself.
            return (0, 0)

        cell_box = self._create_box_for_network(grid)
        connections = self._create_connection_boxes(cell_box, grid)

        # Check if at least one of the connections exists
        valid_connection = connections[0] if connections[0] else connections[1]
        if valid_connection is None:
            return (0, 0)

        # Convenience variables
        v_axis = self.voltage_axis
        l_axis = self.lateral_axis
        n_axis = self.normal_axis
        cell_size = cell_box.size

        # Get common properties of the connections
        grid_centers = grid.centers.to_list[self.normal_axis]
        ub = np.searchsorted(grid_centers, cell_box.center[self.normal_axis])
        thickness_eff = grid_centers[ub] - grid_centers[ub - 1]
        width_eff = valid_connection.size[l_axis]
        # After discretization a wire has an effective width equal to the grid cell size
        if self.size[l_axis] == 0:
            width_eff = cell_size[l_axis]
        # If there are two connections, they will share the same thickness and width
        # only their lengths along the voltage axis might be different
        common_size = list(valid_connection.size)
        common_size[n_axis] = thickness_eff
        common_size[l_axis] = width_eff

        if connections[0] and connections[1]:
            # Typical case of connections above and below network portion
            d_sep = cell_size[v_axis]
            wire_1_size = list(common_size)
            wire_2_size = list(common_size)
            wire_1_size[v_axis] = connections[0].size[v_axis]
            wire_2_size[v_axis] = connections[1].size[v_axis]
            L = total_inductance_colinear_rectangular_wire_segments(
                wire_1_size, wire_2_size, d_sep, v_axis
            )
            # Average length of the two connections
            l_eff = 0.5 * (wire_1_size[v_axis] + wire_2_size[v_axis])
            # Rough equivalent radius based on perimeter
            r_eff = 2 * (width_eff + thickness_eff) / (2 * np.pi)
            approximate_as_wires = width_eff < 4 * thickness_eff and r_eff < l_eff / 4
            if approximate_as_wires:
                C = capacitance_colinear_cylindrical_wire_segments(r_eff, l_eff, d_sep)
            else:
                C = capacitance_rectangular_sheets(width_eff, l_eff, d_sep)
            return (L, C)
        if connections[0] or connections[1]:
            # Possible to only have a single connection, where the capacitance will be 0
            # but there will be a contribution to inductance from the single connection
            L = inductance_straight_rectangular_wire(common_size, v_axis)
            return (L, 0)

    def admittance(self, freqs: np.ndarray) -> np.ndarray:
        """Returns the admittance of this lumped element at the frequencies specified by ``freqs``.

        Note
        ----

        Admittance is returned using the physics convention for time-harmonic fields
        :math:`\\exp{-j \\omega t}`, so the imaginary part of the admittance will have
        an opposite sign compared to the expected value when using the engineering convention.
        """
        a, b = self.network._as_admittance_function
        return NetworkConversions.complex_conductivity(a=a, b=b, freqs=freqs)

    def impedance(self, freqs: np.ndarray) -> np.ndarray:
        """Returns the impedance of this lumped element at the frequencies specified by ``freqs``.

        Note
        ----

        Impedance is returned using the physics convention for time-harmonic fields
        :math:`\\exp{-j \\omega t}`, so the imaginary part of the impedance will have
        an opposite sign compared to the expected value when using the engineering convention.
        """
        return 1.0 / self.admittance(freqs=freqs)


# lumped elements allowed in Simulation.lumped_elements
LumpedElementType = Annotated[
    Union[
        LumpedResistor,
        CoaxialLumpedResistor,
        LinearLumpedElement,
    ],
    pd.Field(discriminator=TYPE_TAG_STR),
]



"""Terminal component modeler analysis functions.

This module contains functions for constructing S-matrices and computing wave amplitudes
for terminal-based component modeling in electromagnetic simulations.

References
----------
.. [1]  R. B. Marks and D. F. Williams, "A general waveguide circuit theory,"
        J. Res. Natl. Inst. Stand. Technol., vol. 97, pp. 533, 1992.

.. [2]  D. M. Pozar, Microwave Engineering, 4th ed. Hoboken, NJ, USA:
        John Wiley & Sons, 2012.
"""

from __future__ import annotations

import numpy as np

from tidy3d.components.data.sim_data import SimulationData
from tidy3d.plugins.smatrix.component_modelers.terminal import TerminalComponentModeler
from tidy3d.plugins.smatrix.data.data_array import PortDataArray, TerminalPortDataArray
from tidy3d.plugins.smatrix.data.terminal import TerminalComponentModelerData
from tidy3d.plugins.smatrix.ports.wave import WavePort
from tidy3d.plugins.smatrix.types import SParamDef
from tidy3d.plugins.smatrix.utils import (
    ab_to_s,
    check_port_impedance_sign,
    compute_F,
    compute_port_VI,
)


def terminal_construct_smatrix(
    modeler_data: TerminalComponentModelerData,
    assume_ideal_excitation: bool = False,
    s_param_def: SParamDef = "pseudo",
) -> TerminalPortDataArray:
    """Constructs the scattering matrix (S-matrix) from raw simulation data.

    This function iterates through each port excitation simulation. For each run,
    it calculates the resulting incident ('a') and reflected ('b') wave
    amplitudes at all ports. These amplitudes are compiled into matrices,
    which are then used to compute the final S-matrix.

    If all ports are excited and ``assume_ideal_excitation`` is ``False``, the
    S-matrix is computed using the formula :math:`S = b a^{-1}`. Otherwise,
    it is assumed that the incident wave matrix 'a' is diagonal, and the
    S-matrix is computed more efficiently by scaling the 'b' matrix. This
    is also necessary when only a subset of ports are excited.

    Parameters
    ----------
    modeler_data : TerminalComponentModelerData
        Data object containing the modeler definition and the raw
        results from each port simulation run.
    assume_ideal_excitation : bool, optional
        If ``True``, assumes that exciting one port does not produce incident
        waves at other ports. This simplifies the S-matrix calculation and is
        required if not all ports are excited. Default is ``False``.
    s_param_def : SParamDef, optional
        The definition of S-parameters to use depends whether "pseudo waves"
        or "power waves" are calculated. Default is "pseudo".

    Returns
    -------
    TerminalPortDataArray
        The computed S-matrix as a :class:`.TerminalPortDataArray` with dimensions
        for frequency, output port, and input port.
    """
    monitor_indices = list(modeler_data.modeler.matrix_indices_monitor)
    source_indices = list(modeler_data.modeler.matrix_indices_source)
    run_source_indices = list(modeler_data.modeler.matrix_indices_run_sim)

    values = np.zeros(
        (len(modeler_data.modeler.freqs), len(monitor_indices), len(source_indices)),
        dtype=complex,
    )
    coords = {
        "f": np.array(modeler_data.modeler.freqs),
        "port_out": monitor_indices,
        "port_in": source_indices,
    }
    a_matrix = TerminalPortDataArray(values, coords=coords)
    b_matrix = a_matrix.copy(deep=True)

    # Tabulate the reference impedances at each port and frequency
    port_impedances = port_reference_impedances(modeler_data=modeler_data)

    for source_index in run_source_indices:
        port, mode_index = modeler_data.modeler.network_dict[source_index]
        sim_data = modeler_data.data[
            modeler_data.modeler.get_task_name(port=port, mode_index=mode_index)
        ]
        a, b = modeler_data.compute_wave_amplitudes_at_each_port(
            port_reference_impedances=port_impedances, sim_data=sim_data, s_param_def=s_param_def
        )

        indexer = {"port_in": source_index}
        a_matrix = a_matrix._with_updated_data(data=a.data, coords=indexer)
        b_matrix = b_matrix._with_updated_data(data=b.data, coords=indexer)

    # If excitation is assumed ideal, a_matrix is assumed to be diagonal
    # and the explicit inverse can be avoided. When only a subset of excitations
    # have been run, we cannot find the inverse anyways so must make this assumption.
    if len(monitor_indices) == len(run_source_indices) and not assume_ideal_excitation:
        s_matrix = ab_to_s(a_matrix, b_matrix)
    else:
        a_diag = np.diagonal(a_matrix, axis1=1, axis2=2)
        # Scale each column by the corresponding diagonal entry
        s_matrix = b_matrix / a_diag[:, np.newaxis, :]

    # element can be determined by user-defined mapping
    for (row_in, col_in), (row_out, col_out), mult_by in modeler_data.modeler.element_mappings:
        coords_from = {
            "port_in": col_in,
            "port_out": row_in,
        }
        coords_to = {
            "port_in": col_out,
            "port_out": row_out,
        }
        data = mult_by * s_matrix.loc[coords_from].data
        s_matrix = s_matrix._with_updated_data(data=data, coords=coords_to)
    return s_matrix


def port_reference_impedances(modeler_data: TerminalComponentModelerData) -> PortDataArray:
    """Calculates the reference impedance for each port across all frequencies.

    This function determines the characteristic impedance for every port defined
    in the modeler. It handles two types of ports differently: for a
    :class:`.WavePort`, the impedance is frequency-dependent and computed from
    modal properties, while for other types like :class:`.LumpedPort`, the
    impedance is a user-defined constant value.

    Parameters
    ----------
    modeler_data : TerminalComponentModelerData
        Data object containing the modeler definition and the raw
        simulation data needed for :class:`.WavePort` impedance calculations.

    Returns
    -------
    PortDataArray
        A ``PortDataArray`` containing the complex impedance for each port at each
        frequency.
    """
    values = np.zeros(
        (len(modeler_data.modeler.freqs), len(modeler_data.modeler.matrix_indices_monitor)),
        dtype=complex,
    )

    coords = {
        "f": np.array(modeler_data.modeler.freqs),
        "port": list(modeler_data.modeler.matrix_indices_monitor),
    }
    port_impedances = PortDataArray(values, coords=coords)
    # Each simulation will store the results from the ModeMonitors,
    # so here we just choose the first one.
    first_sim_index = modeler_data.modeler.matrix_indices_run_sim[0]
    port, mode_index = modeler_data.modeler.network_dict[first_sim_index]
    sim_data = modeler_data.data[
        modeler_data.modeler.get_task_name(port=port, mode_index=mode_index)
    ]
    for network_index in modeler_data.modeler.matrix_indices_monitor:
        port, mode_index = modeler_data.modeler.network_dict[network_index]
        indexer = {"port": network_index}
        if isinstance(port, WavePort):
            # WavePorts have a port impedance calculated from its associated modal field distribution
            # and is frequency dependent.
            data = port.compute_port_impedance(sim_data).data
            port_impedances = port_impedances._with_updated_data(data=data, coords=indexer)
        else:
            # LumpedPorts have a constant reference impedance
            data = np.full(len(modeler_data.modeler.freqs), port.impedance)
            port_impedances = port_impedances._with_updated_data(data=data, coords=indexer)

    port_impedances = modeler_data.modeler._set_port_data_array_attributes(port_impedances)
    return port_impedances


def compute_wave_amplitudes_at_each_port(
    modeler: TerminalComponentModeler,
    port_reference_impedances: PortDataArray,
    sim_data: SimulationData,
    s_param_def: SParamDef = "pseudo",
) -> tuple[PortDataArray, PortDataArray]:
    """Compute the incident and reflected amplitudes at each port.

    The computed amplitudes have not been normalized.

    Parameters
    ----------
    modeler : :class:`.TerminalComponentModeler`
        The component modeler defining the ports and simulation settings.
    port_reference_impedances : :class:`.PortDataArray`
        Reference impedance at each port.
    sim_data : :class:`.SimulationData`
        Results from a single simulation run.
    s_param_def : SParamDef
        The type of waves computed, either pseudo waves defined by Equation 53 and
        Equation 54 in [1], or power waves defined by Equation 4.67 in [2].

    Returns
    -------
    tuple[:class:`.PortDataArray`, :class:`.PortDataArray`]
        Incident (a) and reflected (b) wave amplitudes at each port.
    """
    network_indices = list(modeler.matrix_indices_monitor)
    values = np.zeros(
        (len(modeler.freqs), len(network_indices)),
        dtype=complex,
    )
    coords = {
        "f": np.array(modeler.freqs),
        "port": network_indices,
    }

    V_matrix = PortDataArray(values, coords=coords)
    I_matrix = V_matrix.copy(deep=True)
    a = V_matrix.copy(deep=True)
    b = V_matrix.copy(deep=True)

    for network_index in network_indices:
        port, mode_index = modeler.network_dict[network_index]
        V_out, I_out = compute_port_VI(port, sim_data)
        indexer = {"port": network_index}
        V_matrix = V_matrix._with_updated_data(data=V_out.data, coords=indexer)
        I_matrix = I_matrix._with_updated_data(data=I_out.data, coords=indexer)

    V_numpy = V_matrix.values
    I_numpy = I_matrix.values
    Z_numpy = port_reference_impedances.values

    # Check to make sure sign is consistent for all impedance values
    check_port_impedance_sign(Z_numpy)

    # Check for negative real part of port impedance and flip the V and Z signs accordingly
    negative_real_Z = np.real(Z_numpy) < 0
    V_numpy = np.where(negative_real_Z, -V_numpy, V_numpy)
    Z_numpy = np.where(negative_real_Z, -Z_numpy, Z_numpy)

    F_numpy = compute_F(Z_numpy, s_param_def)

    b_Zref = Z_numpy
    if s_param_def == "power":
        b_Zref = np.conj(Z_numpy)

    # Equations 53 and 54 from [1]
    # Equation 4.67 - Pozar - Microwave Engineering 4ed
    a.values = F_numpy * (V_numpy + Z_numpy * I_numpy)
    b.values = F_numpy * (V_numpy - b_Zref * I_numpy)

    return a, b


def compute_power_wave_amplitudes_at_each_port(
    modeler: TerminalComponentModeler,
    port_reference_impedances: PortDataArray,
    sim_data: SimulationData,
) -> tuple[PortDataArray, PortDataArray]:
    """Compute the incident and reflected power wave amplitudes at each port.

    This is a convenience function that calls
    :func:`.compute_wave_amplitudes_at_each_port` with ``s_param_def="power"``.
    The computed amplitudes have not been normalized.

    Parameters
    ----------
    modeler : :class:`.TerminalComponentModeler`
        The component modeler defining the ports and simulation settings.
    port_reference_impedances : :class:`.PortDataArray`
        Reference impedance at each port.
    sim_data : :class:`.SimulationData`
        Results from a single simulation run.

    Returns
    -------
    tuple[:class:`.PortDataArray`, :class:`.PortDataArray`]
        Incident (a) and reflected (b) power wave amplitudes at each port.
    """
    return compute_wave_amplitudes_at_each_port(
        modeler=modeler,
        port_reference_impedances=port_reference_impedances,
        sim_data=sim_data,
        s_param_def="power",
    )



"""Component imports for adjoint plugin. from tidy3d.plugins.adjoint.components import *"""

# import the jax version of tidy3d components
from __future__ import annotations

from .data.data_array import JaxDataArray
from .data.dataset import JaxPermittivityDataset
from .data.monitor_data import JaxModeData
from .data.sim_data import JaxSimulationData
from .geometry import JaxBox, JaxComplexPolySlab, JaxPolySlab
from .medium import JaxAnisotropicMedium, JaxCustomMedium, JaxMedium
from .simulation import JaxSimulation
from .structure import JaxStructure, JaxStructureStaticGeometry, JaxStructureStaticMedium

__all__ = [
    "JaxAnisotropicMedium",
    "JaxBox",
    "JaxComplexPolySlab",
    "JaxCustomMedium",
    "JaxDataArray",
    "JaxGeometryGroup",
    "JaxMedium",
    "JaxModeData",
    "JaxPermittivityDataset",
    "JaxPolySlab",
    "JaxSimulation",
    "JaxSimulationData",
    "JaxStructure",
    "JaxStructureStaticGeometry",
    "JaxStructureStaticMedium",
]



"""Storing data associated with results from the TerminalComponentModeler"""

from __future__ import annotations

import pydantic.v1 as pd

from tidy3d.components.data.data_array import DataArray
from tidy3d.log import log


class PortDataArray(DataArray):
    """Array of values over dimensions of frequency and port name.

    Example
    -------
    >>> import numpy as np
    >>> f = [2e9, 3e9, 4e9]
    >>> ports = ["port1", "port2"]
    >>> coords = dict(f=f, port=ports)
    >>> data = (1+1j) * np.random.random((3, 2))
    >>> port_data = PortDataArray(data, coords=coords)
    """

    __slots__ = ()
    _dims = ("f", "port")

    @pd.root_validator(pre=False)
    def _warn_rf_license(cls, values):
        log.warning(
            "ℹ️ ⚠️ RF simulations are subject to new license requirements in the future. You have instantiated at least one RF-specific component.",
            log_once=True,
        )
        return values


class ModalPortDataArray(DataArray):
    """Port parameter matrix elements for modal ports.

    Example
    -------
    >>> import numpy as np
    >>> ports_in = ['port1', 'port2']
    >>> ports_out = ['port1', 'port2']
    >>> mode_index_in = [0, 1]
    >>> mode_index_out = [0, 1]
    >>> f = [2e14]
    >>> coords = dict(
    ...     port_in=ports_in,
    ...     port_out=ports_out,
    ...     mode_index_in=mode_index_in,
    ...     mode_index_out=mode_index_out,
    ...     f=f
    ... )
    >>> port_data = ModalPortDataArray((1 + 1j) * np.random.random((2, 2, 2, 2, 1)), coords=coords)
    """

    __slots__ = ()
    _dims = ("port_out", "mode_index_out", "port_in", "mode_index_in", "f")
    _data_attrs = {"long_name": "modal port matrix element"}


class TerminalPortDataArray(DataArray):
    """Port parameter matrix elements for terminal-based ports.

    Example
    -------
    >>> import numpy as np
    >>> ports_in = ["port1", "port2"]
    >>> ports_out = ["port1", "port2"]
    >>> f = [2e14]
    >>> coords = dict(f=f, port_out=ports_out, port_in=ports_in)
    >>> data = (1+1j) * np.random.random((1, 2, 2))
    >>> port_data = TerminalPortDataArray(data, coords=coords)
    """

    __slots__ = ()
    _dims = ("f", "port_out", "port_in")
    _data_attrs = {"long_name": "terminal-based port matrix element"}

    @pd.root_validator(pre=False)
    def _warn_rf_license(cls, values):
        log.warning(
            "ℹ️ ⚠️ RF simulations are subject to new license requirements in the future. You have instantiated at least one RF-specific component.",
            log_once=True,
        )
        return values



"""
This class defines standard SPICE electrical_analysis types (electrical simulations configurations).
"""

from __future__ import annotations

import pydantic.v1 as pd

from tidy3d.components.base import Tidy3dBaseModel
from tidy3d.constants import KELVIN


class ChargeToleranceSpec(Tidy3dBaseModel):
    """
    Charge tolerance parameters relevant to multiple simulation analysis types.

    Example
    -------
    >>> import tidy3d as td
    >>> charge_settings = td.ChargeToleranceSpec(abs_tol=1e8, rel_tol=1e-10, max_iters=30)
    """

    abs_tol: pd.PositiveFloat = pd.Field(
        default=1e10,
        title="Absolute tolerance.",
        description="Absolute tolerance used as stop criteria when converging towards a solution.",
    )

    rel_tol: pd.PositiveFloat = pd.Field(
        default=1e-10,
        title="Relative tolerance.",
        description="Relative tolerance used as stop criteria when converging towards a solution.",
    )

    max_iters: pd.PositiveInt = pd.Field(
        default=30,
        title="Maximum number of iterations.",
        description="Indicates the maximum number of iterations to be run. "
        "The solver will stop either when this maximum of iterations is met "
        "or when the tolerance criteria has been met.",
    )

    ramp_up_iters: pd.PositiveInt = pd.Field(
        default=1,
        title="Ramp-up iterations.",
        description="In order to help in start up, quantities such as doping "
        "are ramped up until they reach their specified value. This parameter "
        "determines how many of this iterations it takes to reach full values.",
    )


class IsothermalSteadyChargeDCAnalysis(Tidy3dBaseModel):
    """
    Configures relevant steady-state DC simulation parameters for a charge simulation.
    """

    temperature: pd.PositiveFloat = pd.Field(
        300,
        title="Temperature",
        description="Lattice temperature. Assumed constant throughout the device. "
        "Carriers are assumed to be at thermodynamic equilibrium with the lattice.",
        units=KELVIN,
    )

    tolerance_settings: ChargeToleranceSpec = pd.Field(
        default=ChargeToleranceSpec(), title="Tolerance settings"
    )

    convergence_dv: pd.PositiveFloat = pd.Field(
        default=1.0,
        title="Bias step.",
        description="By default, a solution is computed at 0 bias. If a bias different than "
        "0 is requested through a voltage source, the charge solver will start at 0 and increase bias "
        "at `convergence_dv` intervals until the required bias is reached. This is, therefore, a "
        "convergence parameter in DC computations.",
    )

    fermi_dirac: bool = pd.Field(
        False,
        title="Fermi-Dirac statistics",
        description="Determines whether Fermi-Dirac statistics are used. When False, "
        "Boltzmann statistics will be used. This can provide more accurate results in situations "
        "where very high doping may lead the pseudo-Fermi energy level to approach "
        "either the conduction or valence energy bands.",
    )



# Defines specifications for how long to run a simulation
from __future__ import annotations

import pydantic.v1 as pd

from .base import Tidy3dBaseModel


class RunTimeSpec(Tidy3dBaseModel):
    """Defines specification for how long to run a simulation when added to ``Simulation.run_time``.

    Notes
    -----

        The evaluated ``run_time`` will be computed from a ``RunTimeSpec()`` as follows:

        .. math::

            \\text{run_time} =  \\text{source_factor} * T_{src_max} + \\text{quality_factor} n_{max} L_{max} / c_{0}

        Where: ``source_factor`` and ``quality_factor`` are fields in the spec,
        :math:`T_{src_max}` is the longest time that a source is non-zero,
        :math:`n_{max}` is the maximum refractive index in the simulation,
        :math:`L_{max}` is the distance along the largest dimension in the simulation, and
        :math:`c_0` is the speed of light in vacuum.

    """

    quality_factor: pd.PositiveFloat = pd.Field(
        ...,
        title="Quality Factor",
        description="Quality factor expected in the device. This determines how long the "
        "simulation will run as it assumes a field decay time that scales proportionally to "
        "this value.",
    )

    source_factor: pd.PositiveFloat = pd.Field(
        3,
        title="Source Factor",
        description="The contribution to the ``run_time`` from the longest source is computed from "
        "the ``source_time`` length times ``source_factor``. Larger values provide more buffer "
        "at the expense of potentially giving ``run_time`` values that are larger than needed.",
    )



"""Class for computing characteristic impedance of transmission lines."""

from __future__ import annotations

from typing import Optional, Union

import numpy as np
import pydantic.v1 as pd

from tidy3d.components.base import Tidy3dBaseModel
from tidy3d.components.data.data_array import ImpedanceResultTypes, _make_impedance_data_array
from tidy3d.components.data.monitor_data import FieldTimeData
from tidy3d.components.monitor import ModeMonitor, ModeSolverMonitor
from tidy3d.exceptions import ValidationError
from tidy3d.log import log

from .custom_path_integrals import CustomCurrentIntegral2D, CustomVoltageIntegral2D
from .path_integrals import (
    AxisAlignedPathIntegral,
    CurrentIntegralAxisAligned,
    MonitorDataTypes,
    VoltageIntegralAxisAligned,
)

VoltageIntegralTypes = Union[VoltageIntegralAxisAligned, CustomVoltageIntegral2D]
CurrentIntegralTypes = Union[CurrentIntegralAxisAligned, CustomCurrentIntegral2D]


class ImpedanceCalculator(Tidy3dBaseModel):
    """Tool for computing the characteristic impedance of a transmission line."""

    voltage_integral: Optional[VoltageIntegralTypes] = pd.Field(
        None,
        title="Voltage Integral",
        description="Definition of path integral for computing voltage.",
    )

    current_integral: Optional[CurrentIntegralTypes] = pd.Field(
        None,
        title="Current Integral",
        description="Definition of contour integral for computing current.",
    )

    def compute_impedance(self, em_field: MonitorDataTypes) -> ImpedanceResultTypes:
        """Compute impedance for the supplied ``em_field`` using ``voltage_integral`` and
        ``current_integral``. If only a single integral has been defined, impedance is
        computed using the total flux in ``em_field``.

        Parameters
        ----------
        em_field : :class:`.MonitorDataTypes`
            The electromagnetic field data that will be used for computing the characteristic
            impedance.

        Returns
        -------
        :class:`.ImpedanceResultTypes`
            Result of impedance computation over remaining dimensions (frequency, time, mode indices).
        """

        AxisAlignedPathIntegral._check_monitor_data_supported(em_field=em_field)

        # If both voltage and current integrals have been defined then impedance is computed directly
        if self.voltage_integral is not None:
            voltage = self.voltage_integral.compute_voltage(em_field)
        if self.current_integral is not None:
            current = self.current_integral.compute_current(em_field)

        # If only one of the integrals has been provided, then the computation falls back to using
        # total power (flux) with Ohm's law to compute the missing quantity. The input field should
        # cover an area large enough to render the flux computation accurate. If the input field is
        # a time signal, then it is real and flux corresponds to the instantaneous power. Otherwise
        # the input field is in frequency domain, where flux indicates the time-averaged power
        # 0.5*Re(V*conj(I)).
        # We explicitly take the real part, in case Bloch BCs were used in the simulation.
        flux_sign = 1.0
        # Determine flux sign
        if isinstance(em_field.monitor, ModeSolverMonitor):
            flux_sign = 1 if em_field.monitor.direction == "+" else -1
        if isinstance(em_field.monitor, ModeMonitor):
            flux_sign = 1 if em_field.monitor.store_fields_direction == "+" else -1

        if self.voltage_integral is None:
            flux = flux_sign * em_field.complex_flux
            if isinstance(em_field, FieldTimeData):
                impedance = flux / np.real(current) ** 2
            else:
                impedance = 2 * flux / (current * np.conj(current))
        elif self.current_integral is None:
            flux = flux_sign * em_field.complex_flux
            if isinstance(em_field, FieldTimeData):
                impedance = np.real(voltage) ** 2 / flux
            else:
                impedance = (voltage * np.conj(voltage)) / (2 * np.conj(flux))
        else:
            if isinstance(em_field, FieldTimeData):
                impedance = np.real(voltage) / np.real(current)
            else:
                impedance = voltage / current
        impedance = _make_impedance_data_array(impedance)
        return impedance

    @pd.validator("current_integral", always=True)
    def check_voltage_or_current(cls, val, values):
        """Raise validation error if both ``voltage_integral`` and ``current_integral``
        are not provided."""
        if not values.get("voltage_integral") and not val:
            raise ValidationError(
                "At least one of 'voltage_integral' or 'current_integral' must be provided."
            )
        return val

    @pd.root_validator(pre=False)
    def _warn_rf_license(cls, values):
        log.warning(
            "ℹ️ ⚠️ RF simulations are subject to new license requirements in the future. You have instantiated at least one RF-specific component.",
            log_once=True,
        )
        return values



# utilities for working with autograd
from __future__ import annotations

from collections.abc import Iterable
from typing import Any

import autograd.numpy as anp
from autograd.tracer import getval

__all__ = [
    "asarray1d",
    "contains",
    "get_static",
    "is_tidy_box",
    "pack_complex_vec",
    "split_list",
]


def get_static(x: Any) -> Any:
    """Get the 'static' (untraced) version of some value."""
    return getval(x)


def split_list(x: list[Any], index: int) -> (list[Any], list[Any]):
    """Split a list at a given index."""
    x = list(x)
    return x[:index], x[index:]


def is_tidy_box(x: Any) -> bool:
    """Check if a value is a tidy box."""
    return getattr(x, "_tidy", False)


def contains(target: Any, seq: Iterable[Any]) -> bool:
    """Return ``True`` if target occurs anywhere within arbitrarily nested iterables."""
    for x in seq:
        if x == target:
            return True
        if isinstance(x, Iterable) and not isinstance(x, (str, bytes)):
            if contains(target, x):
                return True
    return False


def pack_complex_vec(z):
    """Ravel [Re(z); Im(z)] into one real vector (autograd-safe)."""
    return anp.concatenate([anp.ravel(anp.real(z)), anp.ravel(anp.imag(z))])


def asarray1d(x):
    """Autograd-friendly 1D flatten: returns ndarray of shape (-1,)."""
    x = anp.array(x)
    return x if x.ndim == 1 else anp.ravel(x)



"""Abstract base for defining simulation classes of different solvers"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Optional

import autograd.numpy as anp
import pydantic.v1 as pd

from tidy3d.components.base import cached_property, skip_if_fields_missing
from tidy3d.components.geometry.base import Box
from tidy3d.components.medium import Medium, MediumType3D
from tidy3d.components.scene import Scene
from tidy3d.components.structure import Structure
from tidy3d.components.types import (
    TYPE_TAG_STR,
    Ax,
    Axis,
    Bound,
    LengthUnit,
    PriorityMode,
    Symmetry,
)
from tidy3d.components.validators import (
    _warn_unsupported_traced_argument,
    assert_objects_in_sim_bounds,
    assert_unique_names,
)
from tidy3d.components.viz import PlotParams, add_ax_if_none, equal_aspect, plot_params_symmetry
from tidy3d.exceptions import Tidy3dKeyError
from tidy3d.log import log
from tidy3d.version import __version__

from .monitor import AbstractMonitor


class AbstractSimulation(Box, ABC):
    """Base class for simulation classes of different solvers."""

    medium: MediumType3D = pd.Field(
        Medium(),
        title="Background Medium",
        description="Background medium of simulation, defaults to vacuum if not specified.",
        discriminator=TYPE_TAG_STR,
    )
    """
    Background medium of simulation, defaults to vacuum if not specified.
    """

    structures: tuple[Structure, ...] = pd.Field(
        (),
        title="Structures",
        description="Tuple of structures present in simulation. "
        "Note: Structures defined later in this list override the "
        "simulation material properties in regions of spatial overlap.",
    )
    """
    Tuple of structures present in simulation. Structures defined later in this list override the simulation
    material properties in regions of spatial overlap.

    Example
    -------
    Simple application reference:

    .. code-block:: python

        Simulation(
            ...
            structures=[
                 Structure(
                 geometry=Box(size=(1, 1, 1), center=(0, 0, 0)),
                 medium=Medium(permittivity=2.0),
                 ),
            ],
            ...
        )
    """

    symmetry: tuple[Symmetry, Symmetry, Symmetry] = pd.Field(
        (0, 0, 0),
        title="Symmetries",
        description="Tuple of integers defining reflection symmetry across a plane "
        "bisecting the simulation domain normal to the x-, y-, and z-axis "
        "at the simulation center of each axis, respectively. ",
    )

    sources: tuple[None, ...] = pd.Field(
        (),
        title="Sources",
        description="Sources in the simulation.",
    )

    boundary_spec: None = pd.Field(
        None,
        title="Boundaries",
        description="Specification of boundary conditions.",
    )

    monitors: tuple[None, ...] = pd.Field(
        (),
        title="Monitors",
        description="Monitors in the simulation. ",
    )

    grid_spec: None = pd.Field(
        None,
        title="Grid Specification",
        description="Specifications for the simulation grid.",
    )

    version: str = pd.Field(
        __version__,
        title="Version",
        description="String specifying the front end version number.",
    )

    plot_length_units: Optional[LengthUnit] = pd.Field(
        "μm",
        title="Plot Units",
        description="When set to a supported ``LengthUnit``, "
        "plots will be produced with proper scaling of axes and "
        "include the desired unit specifier in labels.",
    )

    structure_priority_mode: PriorityMode = pd.Field(
        "equal",
        title="Structure Priority Setting",
        description="This field only affects structures of `priority=None`. "
        "If `equal`, the priority of those structures is set to 0; if `conductor`, "
        "the priority of structures made of `LossyMetalMedium` is set to 90, "
        "`PECMedium` to 100, and others to 0.",
    )

    """ Validating setup """

    @pd.root_validator(pre=True)
    def _update_simulation(cls, values):
        """Update the simulation if it is an earlier version."""

        # dummy upgrade of version number
        # this should be overriden by each simulation class if needed
        current_version = values.get("version")
        if current_version != __version__ and current_version is not None:
            log.warning(f"updating {cls.__name__} from {current_version} to {__version__}")
            values["version"] = __version__
        return values

    # make sure all names are unique
    _unique_monitor_names = assert_unique_names("monitors")
    _unique_structure_names = assert_unique_names("structures")
    _unique_source_names = assert_unique_names("sources")

    _monitors_in_bounds = assert_objects_in_sim_bounds("monitors", strict_inequality=True)
    _structures_in_bounds = assert_objects_in_sim_bounds("structures", error=False)

    _warn_traced_center = _warn_unsupported_traced_argument("center")
    _warn_traced_size = _warn_unsupported_traced_argument("size")

    @pd.validator("structures", always=True)
    @skip_if_fields_missing(["size", "center"])
    def _structures_not_at_edges(cls, val, values):
        """Warn if any structures lie at the simulation boundaries."""

        if val is None:
            return val

        sim_box = Box(size=values.get("size"), center=values.get("center"))
        sim_bound_min, sim_bound_max = sim_box.bounds
        sim_bounds = list(sim_bound_min) + list(sim_bound_max)

        with log as consolidated_logger:
            for istruct, structure in enumerate(val):
                struct_bound_min, struct_bound_max = structure.geometry.bounds
                struct_bounds = list(struct_bound_min) + list(struct_bound_max)

                for sim_val, struct_val in zip(sim_bounds, struct_bounds):
                    if anp.isclose(sim_val, struct_val):
                        consolidated_logger.warning(
                            f"Structure at 'structures[{istruct}]' has bounds that extend exactly "
                            "to simulation edges. This can cause unexpected behavior. "
                            "If intending to extend the structure to infinity along one dimension, "
                            "use td.inf as a size variable instead to make this explicit.",
                            custom_loc=["structures", istruct],
                        )
                        continue

        return val

    """ Post-init validators """

    def _post_init_validators(self) -> None:
        """Call validators taking z`self` that get run after init."""
        _ = self.scene

    def validate_pre_upload(self) -> None:
        """Validate the fully initialized simulation is ok for upload to our servers."""

    """ Accounting """

    @cached_property
    def scene(self) -> Scene:
        """Scene instance associated with the simulation."""

        return Scene(
            medium=self.medium,
            structures=self.structures,
            plot_length_units=self.plot_length_units,
            structure_priority_mode=self.structure_priority_mode,
        )

    def get_monitor_by_name(self, name: str) -> AbstractMonitor:
        """Return monitor named 'name'."""
        for monitor in self.monitors:
            if monitor.name == name:
                return monitor
        raise Tidy3dKeyError(f"No monitor named '{name}'")

    @cached_property
    def simulation_bounds(self) -> Bound:
        """Simulation bounds including auxiliary boundary zones such as PML layers."""
        # in this default implementation we just take self.bounds
        # this should be changed in different solvers depending on whether automatic extensions
        # (like pml) are present
        return self.bounds

    @cached_property
    def simulation_geometry(self) -> Box:
        """The entire simulation domain including auxiliary boundary zones such as PML layers.
        It is identical to ``Simulation.geometry`` in the absence of such auxiliary zones.
        """
        rmin, rmax = self.simulation_bounds
        return Box.from_bounds(rmin=rmin, rmax=rmax)

    @cached_property
    def simulation_structure(self) -> Structure:
        """Returns structure representing the domain of the simulation. This differs from
        ``Simulation.scene.background_structure`` in that it has finite extent."""
        return Structure(geometry=self.simulation_geometry, medium=self.medium)

    @equal_aspect
    @add_ax_if_none
    def plot(
        self,
        x: Optional[float] = None,
        y: Optional[float] = None,
        z: Optional[float] = None,
        ax: Ax = None,
        source_alpha: Optional[float] = None,
        monitor_alpha: Optional[float] = None,
        hlim: Optional[tuple[float, float]] = None,
        vlim: Optional[tuple[float, float]] = None,
        fill_structures: bool = True,
        **patch_kwargs,
    ) -> Ax:
        """Plot each of simulation's components on a plane defined by one nonzero x,y,z coordinate.

        Parameters
        ----------
        x : float = None
            position of plane in x direction, only one of x, y, z must be specified to define plane.
        y : float = None
            position of plane in y direction, only one of x, y, z must be specified to define plane.
        z : float = None
            position of plane in z direction, only one of x, y, z must be specified to define plane.
        source_alpha : float = None
            Opacity of the sources. If ``None``, uses Tidy3d default.
        monitor_alpha : float = None
            Opacity of the monitors. If ``None``, uses Tidy3d default.
        ax : matplotlib.axes._subplots.Axes = None
            Matplotlib axes to plot on, if not specified, one is created.
        hlim : Tuple[float, float] = None
            The x range if plotting on xy or xz planes, y range if plotting on yz plane.
        vlim : Tuple[float, float] = None
            The z range if plotting on xz or yz planes, y plane if plotting on xy plane.
        fill_structures : bool = True
            Whether to fill structures with color or just draw outlines.
        Returns
        -------
        matplotlib.axes._subplots.Axes
            The supplied or created matplotlib axes.
        """

        hlim, vlim = Scene._get_plot_lims(
            bounds=self.simulation_bounds, x=x, y=y, z=z, hlim=hlim, vlim=vlim
        )

        ax = self.scene.plot_structures(
            ax=ax, x=x, y=y, z=z, hlim=hlim, vlim=vlim, fill=fill_structures
        )
        ax = self.plot_sources(ax=ax, x=x, y=y, z=z, hlim=hlim, vlim=vlim, alpha=source_alpha)
        ax = self.plot_monitors(ax=ax, x=x, y=y, z=z, hlim=hlim, vlim=vlim, alpha=monitor_alpha)
        ax = Scene._set_plot_bounds(
            bounds=self.simulation_bounds, ax=ax, x=x, y=y, z=z, hlim=hlim, vlim=vlim
        )
        ax = self.plot_boundaries(ax=ax, x=x, y=y, z=z)
        ax = self.plot_symmetries(ax=ax, x=x, y=y, z=z, hlim=hlim, vlim=vlim)

        # Add the default axis labels, tick labels, and title
        ax = Box.add_ax_labels_and_title(
            ax=ax, x=x, y=y, z=z, plot_length_units=self.plot_length_units
        )

        return ax

    @equal_aspect
    @add_ax_if_none
    def plot_sources(
        self,
        x: Optional[float] = None,
        y: Optional[float] = None,
        z: Optional[float] = None,
        hlim: Optional[tuple[float, float]] = None,
        vlim: Optional[tuple[float, float]] = None,
        alpha: Optional[float] = None,
        ax: Ax = None,
    ) -> Ax:
        """Plot each of simulation's sources on a plane defined by one nonzero x,y,z coordinate.

        Parameters
        ----------
        x : float = None
            position of plane in x direction, only one of x, y, z must be specified to define plane.
        y : float = None
            position of plane in y direction, only one of x, y, z must be specified to define plane.
        z : float = None
            position of plane in z direction, only one of x, y, z must be specified to define plane.
        hlim : Tuple[float, float] = None
            The x range if plotting on xy or xz planes, y range if plotting on yz plane.
        vlim : Tuple[float, float] = None
            The z range if plotting on xz or yz planes, y plane if plotting on xy plane.
        alpha : float = None
            Opacity of the sources, If ``None`` uses Tidy3d default.
        ax : matplotlib.axes._subplots.Axes = None
            Matplotlib axes to plot on, if not specified, one is created.

        Returns
        -------
        matplotlib.axes._subplots.Axes
            The supplied or created matplotlib axes.
        """
        bounds = self.bounds
        for source in self.sources:
            ax = source.plot(x=x, y=y, z=z, alpha=alpha, ax=ax, sim_bounds=bounds)
        ax = Scene._set_plot_bounds(
            bounds=self.simulation_bounds, ax=ax, x=x, y=y, z=z, hlim=hlim, vlim=vlim
        )
        # Add the default axis labels, tick labels, and title
        ax = Box.add_ax_labels_and_title(
            ax=ax, x=x, y=y, z=z, plot_length_units=self.plot_length_units
        )
        return ax

    @equal_aspect
    @add_ax_if_none
    def plot_monitors(
        self,
        x: Optional[float] = None,
        y: Optional[float] = None,
        z: Optional[float] = None,
        hlim: Optional[tuple[float, float]] = None,
        vlim: Optional[tuple[float, float]] = None,
        alpha: Optional[float] = None,
        ax: Ax = None,
    ) -> Ax:
        """Plot each of simulation's monitors on a plane defined by one nonzero x,y,z coordinate.

        Parameters
        ----------
        x : float = None
            position of plane in x direction, only one of x, y, z must be specified to define plane.
        y : float = None
            position of plane in y direction, only one of x, y, z must be specified to define plane.
        z : float = None
            position of plane in z direction, only one of x, y, z must be specified to define plane.
        hlim : Tuple[float, float] = None
            The x range if plotting on xy or xz planes, y range if plotting on yz plane.
        vlim : Tuple[float, float] = None
            The z range if plotting on xz or yz planes, y plane if plotting on xy plane.
        alpha : float = None
            Opacity of the sources, If ``None`` uses Tidy3d default.
        ax : matplotlib.axes._subplots.Axes = None
            Matplotlib axes to plot on, if not specified, one is created.

        Returns
        -------
        matplotlib.axes._subplots.Axes
            The supplied or created matplotlib axes.
        """
        bounds = self.bounds
        for monitor in self.monitors:
            ax = monitor.plot(x=x, y=y, z=z, alpha=alpha, ax=ax, sim_bounds=bounds)
        ax = Scene._set_plot_bounds(
            bounds=self.simulation_bounds, ax=ax, x=x, y=y, z=z, hlim=hlim, vlim=vlim
        )
        # Add the default axis labels, tick labels, and title
        ax = Box.add_ax_labels_and_title(
            ax=ax, x=x, y=y, z=z, plot_length_units=self.plot_length_units
        )
        return ax

    @equal_aspect
    @add_ax_if_none
    def plot_symmetries(
        self,
        x: Optional[float] = None,
        y: Optional[float] = None,
        z: Optional[float] = None,
        hlim: Optional[tuple[float, float]] = None,
        vlim: Optional[tuple[float, float]] = None,
        ax: Ax = None,
    ) -> Ax:
        """Plot each of simulation's symmetries on a plane defined by one nonzero x,y,z coordinate.

        Parameters
        ----------
        x : float = None
            position of plane in x direction, only one of x, y, z must be specified to define plane.
        y : float = None
            position of plane in y direction, only one of x, y, z must be specified to define plane.
        z : float = None
            position of plane in z direction, only one of x, y, z must be specified to define plane.
        hlim : Tuple[float, float] = None
            The x range if plotting on xy or xz planes, y range if plotting on yz plane.
        vlim : Tuple[float, float] = None
            The z range if plotting on xz or yz planes, y plane if plotting on xy plane.
        ax : matplotlib.axes._subplots.Axes = None
            Matplotlib axes to plot on, if not specified, one is created.

        Returns
        -------
        matplotlib.axes._subplots.Axes
            The supplied or created matplotlib axes.
        """

        normal_axis, _ = Box.parse_xyz_kwargs(x=x, y=y, z=z)

        for sym_axis, sym_value in enumerate(self.symmetry):
            if sym_value == 0 or sym_axis == normal_axis:
                continue
            sym_box = self._make_symmetry_box(sym_axis=sym_axis)
            plot_params = self._make_symmetry_plot_params(sym_value=sym_value)
            ax = sym_box.plot(x=x, y=y, z=z, ax=ax, **plot_params.to_kwargs())
        ax = Scene._set_plot_bounds(
            bounds=self.simulation_bounds, ax=ax, x=x, y=y, z=z, hlim=hlim, vlim=vlim
        )
        # Add the default axis labels, tick labels, and title
        ax = Box.add_ax_labels_and_title(
            ax=ax, x=x, y=y, z=z, plot_length_units=self.plot_length_units
        )
        return ax

    def _make_symmetry_plot_params(self, sym_value: Symmetry) -> PlotParams:
        """Make PlotParams for symmetry."""

        plot_params = plot_params_symmetry.copy()

        if sym_value == 1:
            plot_params = plot_params.copy(
                update={"facecolor": "lightsteelblue", "edgecolor": "lightsteelblue", "hatch": "++"}
            )
        elif sym_value == -1:
            plot_params = plot_params.copy(
                update={"facecolor": "goldenrod", "edgecolor": "goldenrod", "hatch": "--"}
            )

        return plot_params

    def _make_symmetry_box(self, sym_axis: Axis) -> Box:
        """Construct a :class:`.Box` representing the symmetry to be plotted."""
        rmin, rmax = (list(bound) for bound in self.simulation_bounds)
        rmax[sym_axis] = (rmin[sym_axis] + rmax[sym_axis]) / 2

        return Box.from_bounds(rmin, rmax)

    @abstractmethod
    @equal_aspect
    @add_ax_if_none
    def plot_boundaries(
        self,
        x: Optional[float] = None,
        y: Optional[float] = None,
        z: Optional[float] = None,
        ax: Ax = None,
        **kwargs,
    ) -> Ax:
        """Plot the simulation boundary conditions as lines on a plane
           defined by one nonzero x,y,z coordinate.

        Parameters
        ----------
        x : float = None
            position of plane in x direction, only one of x, y, z must be specified to define plane.
        y : float = None
            position of plane in y direction, only one of x, y, z must be specified to define plane.
        z : float = None
            position of plane in z direction, only one of x, y, z must be specified to define plane.
        ax : matplotlib.axes._subplots.Axes = None
            Matplotlib axes to plot on, if not specified, one is created.
        **kwargs
            Optional keyword arguments passed to the matplotlib ``LineCollection``.
            For details on accepted values, refer to
            `Matplotlib's documentation <https://tinyurl.com/2p97z4cn>`_.

        Returns
        -------
        matplotlib.axes._subplots.Axes
            The supplied or created matplotlib axes.
        """

    @equal_aspect
    @add_ax_if_none
    def plot_structures(
        self,
        x: Optional[float] = None,
        y: Optional[float] = None,
        z: Optional[float] = None,
        ax: Ax = None,
        hlim: Optional[tuple[float, float]] = None,
        vlim: Optional[tuple[float, float]] = None,
        fill: bool = True,
    ) -> Ax:
        """Plot each of simulation's structures on a plane defined by one nonzero x,y,z coordinate.

        Parameters
        ----------
        x : float = None
            position of plane in x direction, only one of x, y, z must be specified to define plane.
        y : float = None
            position of plane in y direction, only one of x, y, z must be specified to define plane.
        z : float = None
            position of plane in z direction, only one of x, y, z must be specified to define plane.
        ax : matplotlib.axes._subplots.Axes = None
            Matplotlib axes to plot on, if not specified, one is created.
        hlim : Tuple[float, float] = None
            The x range if plotting on xy or xz planes, y range if plotting on yz plane.
        vlim : Tuple[float, float] = None
            The z range if plotting on xz or yz planes, y plane if plotting on xy plane.
        fill : bool = True
            Whether to fill structures with color or just draw outlines.
        Returns
        -------
        matplotlib.axes._subplots.Axes
            The supplied or created matplotlib axes.
        """

        hlim_new, vlim_new = Scene._get_plot_lims(
            bounds=self.simulation_bounds, x=x, y=y, z=z, hlim=hlim, vlim=vlim
        )

        return self.scene.plot_structures(
            x=x, y=y, z=z, ax=ax, hlim=hlim_new, vlim=vlim_new, fill=fill
        )

    @equal_aspect
    @add_ax_if_none
    def plot_structures_eps(
        self,
        x: Optional[float] = None,
        y: Optional[float] = None,
        z: Optional[float] = None,
        freq: Optional[float] = None,
        alpha: Optional[float] = None,
        cbar: bool = True,
        reverse: bool = False,
        ax: Ax = None,
        hlim: Optional[tuple[float, float]] = None,
        vlim: Optional[tuple[float, float]] = None,
    ) -> Ax:
        """Plot each of simulation's structures on a plane defined by one nonzero x,y,z coordinate.
        The permittivity is plotted in grayscale based on its value at the specified frequency.

        Parameters
        ----------
        x : float = None
            position of plane in x direction, only one of x, y, z must be specified to define plane.
        y : float = None
            position of plane in y direction, only one of x, y, z must be specified to define plane.
        z : float = None
            position of plane in z direction, only one of x, y, z must be specified to define plane.
        freq : float = None
            Frequency to evaluate the relative permittivity of all mediums.
            If not specified, evaluates at infinite frequency.
        reverse : bool = False
            If ``False``, the highest permittivity is plotted in black.
            If ``True``, it is plotteed in white (suitable for black backgrounds).
        cbar : bool = True
            Whether to plot a colorbar for the relative permittivity.
        alpha : float = None
            Opacity of the structures being plotted.
            Defaults to the structure default alpha.
        ax : matplotlib.axes._subplots.Axes = None
            Matplotlib axes to plot on, if not specified, one is created.
        hlim : Tuple[float, float] = None
            The x range if plotting on xy or xz planes, y range if plotting on yz plane.
        vlim : Tuple[float, float] = None
            The z range if plotting on xz or yz planes, y plane if plotting on xy plane.

        Returns
        -------
        matplotlib.axes._subplots.Axes
            The supplied or created matplotlib axes.
        """

        hlim, vlim = Scene._get_plot_lims(
            bounds=self.simulation_bounds, x=x, y=y, z=z, hlim=hlim, vlim=vlim
        )

        return self.scene.plot_structures_eps(
            freq=freq,
            cbar=cbar,
            alpha=alpha,
            ax=ax,
            x=x,
            y=y,
            z=z,
            hlim=hlim,
            vlim=vlim,
            reverse=reverse,
        )

    @equal_aspect
    @add_ax_if_none
    def plot_structures_heat_conductivity(
        self,
        x: Optional[float] = None,
        y: Optional[float] = None,
        z: Optional[float] = None,
        alpha: Optional[float] = None,
        cbar: bool = True,
        reverse: bool = False,
        ax: Ax = None,
        hlim: Optional[tuple[float, float]] = None,
        vlim: Optional[tuple[float, float]] = None,
    ) -> Ax:
        """Plot each of simulation's structures on a plane defined by one nonzero x,y,z coordinate.
        The permittivity is plotted in grayscale based on its value at the specified frequency.

        Parameters
        ----------
        x : float = None
            position of plane in x direction, only one of x, y, z must be specified to define plane.
        y : float = None
            position of plane in y direction, only one of x, y, z must be specified to define plane.
        z : float = None
            position of plane in z direction, only one of x, y, z must be specified to define plane.
        freq : float = None
            Frequency to evaluate the relative permittivity of all mediums.
            If not specified, evaluates at infinite frequency.
        reverse : bool = False
            If ``False``, the highest permittivity is plotted in black.
            If ``True``, it is plotteed in white (suitable for black backgrounds).
        cbar : bool = True
            Whether to plot a colorbar for the relative permittivity.
        alpha : float = None
            Opacity of the structures being plotted.
            Defaults to the structure default alpha.
        ax : matplotlib.axes._subplots.Axes = None
            Matplotlib axes to plot on, if not specified, one is created.
        hlim : Tuple[float, float] = None
            The x range if plotting on xy or xz planes, y range if plotting on yz plane.
        vlim : Tuple[float, float] = None
            The z range if plotting on xz or yz planes, y plane if plotting on xy plane.

        Returns
        -------
        matplotlib.axes._subplots.Axes
            The supplied or created matplotlib axes.
        """

        hlim, vlim = Scene._get_plot_lims(
            bounds=self.simulation_bounds, x=x, y=y, z=z, hlim=hlim, vlim=vlim
        )

        return self.scene.plot_structures_heat_conductivity(
            cbar=cbar,
            alpha=alpha,
            ax=ax,
            x=x,
            y=y,
            z=z,
            hlim=hlim,
            vlim=vlim,
            reverse=reverse,
        )

    @classmethod
    def from_scene(cls, scene: Scene, **kwargs) -> AbstractSimulation:
        """Create a simulation from a :class:`.Scene` instance. Must provide additional parameters
        to define a valid simulation (for example, ``size``, ``run_time``, ``grid_spec``, etc).

        Parameters
        ----------
        scene : :class:`.Scene`
            Scene containing structures information.
        **kwargs
            Other arguments
        """

        return cls(
            structures=scene.structures,
            medium=scene.medium,
            **kwargs,
        )
